{-# OPTIONS --sized-types #-}
module Gcd where

postulate
  Size : Set
  ↑_   : Size → Size
  ∞    : Size
  Size< : ..(_ : Size) → Set
  max   : Size → Size → Size

{-# BUILTIN SIZE    Size  #-}
{-# BUILTIN SIZESUC ↑_    #-}
{-# BUILTIN SIZEINF ∞    #-}
{-# BUILTIN SIZELT  Size< #-}
{-# BUILTIN SIZEMAX max   #-}

data Nat {size : Size} : Set where
  zero : Nat
  suc  : {size' : Size< size} → Nat {size'} → Nat

data Either (A B : Set) : Set where
  left  : A → Either A B
  right : B → Either A B

-- symmetric difference: returns which number was bigger, plus the difference
diff : {i j : Size} → Nat {i} → Nat {j} → Either (Nat {i}) (Nat {j})
diff zero    m       = right m
diff (suc n) zero    = left (suc n)
diff (suc n) (suc m) = diff n m

gcd : {i j : Size} → Nat {i} → Nat {j} → Nat {max i j}
gcd         zero    m            = m
gcd         (suc n) zero         = suc n
gcd {j = j} (suc n) (suc {j'} m) with diff {j = j'} n m
... | left n'  = gcd {j = j} n' (suc m)
... | right m' = gcd {j = j'} (suc n) m'

{-
data Nat : {size : Size} → Set where
  zero : {size : Size} → Nat {↑ size}
  suc  : {size : Size} → Nat {size} → Nat {↑ size}
-}
